
/*****************************************************************************/
/*                                                                           */
/*  Practica de Sistemes Informatics en Temps Real                           */
/*                                                                           */
/*  Implementacio d'un sequenciador de MIDI per mitja de Real Time Linux     */
/*  rt_linux.c: fitxer que conte el modul rt_linux amb les RT Tasques i      */
/*              els manegadors de les RT-FIFOs, amb la implementacio de la   */
/*              genereracio de les frequencies MIDI en temps real.           */
/*              S'admet la reproduccio pel Conversor Digital-Analogic o be   */
/*              directament per l'speaker.                                   */ 
/*                                                                           */
/*  Realitzat per: Francisco Jose Aguilar Celdran i Sergio Blanco Cuaresma   */ 
/*                                                                           */
/*****************************************************************************/


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>
#include <asm/io.h>
#include "control.h"

struct taula_musical taula_midi;         // Conte la frecuencia i els tics de cada MIDI
int speaker;                             // 1-> Sona el Speaker ; 0 -> Sona el DAC
RT_TASK tasca_music[2];                  // RT-Tasca music per a flanc ascendent i descendent
RT_TASK tasca_music_speaker;             // RT-Tasca per a music_speaker 
RT_TASK tasca_director;                            // RT-Tasca directora
struct struct_missatge buffers[NUM_BUFFERS];       // Enviat des de linux (tics i duracio)
int buffer_actual;                                 // Indica quin buffer s'empra actualment
int pos_lectura_buffer;                            // Posicio del buffer accedida actualment
int tics_nota_actual;                              // Tics per al periode del MIDI actual
int inici_melodia;                                 // Inidica si s'ha iniciat la melodia
int speaker_tancat;                                // Boolea per coneixer l'estat del speaker
int freq_midi_inicialitzat = 0;                    // Recepcio de frequencies i tics MIDI

/* ------------------------------------------------------------------------------------- */
/*  music: tasca encarregada de realitzar els flancs de pujada o de baixada, en funcio   */
/*         del parametre d'entrada 't', amb el periode en que la tasca directora l'hagi  */
/*         programat per tal de reproduir el MIDI requerit, escribint directament al     */
/*         Conversor Digital-Analogic (DAC).                                             */
/* ------------------------------------------------------------------------------------- */

void music(int t) 
{
   while (1) {
      outb(t, OUTL);         // Escriu a la part baixa del port DAC
      outb(t, OUTH);         // Escriu a la part alta del port DAC
      rt_task_wait();
   }
}

/* -------------------------------------------------------------------------------- */
/*  music_speaker: tasca de similar proposit a la tasca music, que s'encarrega, en  */
/*                 el cas de que hagi estat activat el mode speaker, de reproduir   */
/*                 per l'speaker la frequencia de la posicio actual del buffer.     */
/* -------------------------------------------------------------------------------- */

void music_speaker(int t)
{   
   int pos_midi, freq;

   while (1) {
         pos_midi = buffers[buffer_actual].midi[pos_lectura_buffer];
         if (pos_midi != 0) {
            freq = taula_midi.freq[pos_midi - NUM_PRIMER_MIDI];
         }
         else {
            freq = 0;
         }
         outb(freq &0xff, 0x42);
         outb((freq >> 8) & 0xff, 0x42);
      rt_task_wait();
   }
}

/* ------------------------------------------------------------------------------- */
/*  obrir_speaker: prepara el speaker per poder enviar-hi frequencies a reproduir. */
/* ------------------------------------------------------------------------------- */

void obrir_speaker(void)
{
   outb(inb_p (0xB6) | 3, 0x61);
   outb(0xB6, 0x43);
}

/* ----------------------------------------------- */
/*  tancar_speaker: tanca la sortida del speaker.  */
/* ----------------------------------------------- */

void tancar_speaker(void)
{
   outb(inb_p (0x61) & 0xFC, 0x61);  
}

/* ------------------------------------------------------------------------------- */
/*  inicialitzar_dades: funcio encarregada d'establir els valors inicials a totes  */
/*                      les variables que requereixen previa inicialitzacio abans  */
/*                      de executar a la reproduccio de la melodia.                */
/* ------------------------------------------------------------------------------- */

void inicialitzar_dades(void)
{
   speaker = 0;
   speaker_tancat = 0;
   buffer_actual = 1; // El primer bloc que rep el manegador sera guardat al 0 realment
   inici_melodia = 1;
   pos_lectura_buffer = 0;
}

/* ---------------------------------------------------------------------------------- */
/*  director: tasca directora que dirigeix la reproduccio, programant la tasca music  */
/*            en la frequencia especifica de cada nota MIDI, i enviant peticions de   */
/*            nou buffer en esgotar el buffer actual de notes, emprant la tecnica de  */
/*            doble buffer donat que es reproduiran melodies de tamany desconegut.    */   
/* ---------------------------------------------------------------------------------- */

void director(int t) 
{
   int pos_midi;
   int tics_duracio;
   int tics_flanc_nota;
   int acabar = 0;
   RTIME now;

   while (1) 
   {
      if (pos_lectura_buffer == TAM_BUFFER) {
         pos_lectura_buffer = 0;
         buffer_actual = (buffer_actual + 1) % NUM_BUFFERS;
         if (buffers[buffer_actual].midi[pos_lectura_buffer] == -1) {
            acabar = 1;
         }
         else {
            rtf_put(3, "MES_NOTES\0", 10);
         }
      } 
      else {
         if (buffers[buffer_actual].midi[pos_lectura_buffer] == -1) {
            acabar = 1;
         }
      }
      if (acabar) {
         if (speaker) {
            rt_task_suspend(&tasca_music_speaker);
            tancar_speaker();
            inicialitzar_dades();
            acabar = 0;
            rt_task_suspend(&tasca_director);
         }
         else {
            rt_task_suspend(&tasca_music[0]);
            rt_task_suspend(&tasca_music[1]);
            inicialitzar_dades();
            acabar = 0;
            rt_task_suspend(&tasca_director);
         }
      }
     
      pos_midi = buffers[buffer_actual].midi[pos_lectura_buffer];
      tics_duracio = buffers[buffer_actual].tics[pos_lectura_buffer];

      if (!speaker) {
         if (pos_midi != 0) {
            tics_flanc_nota = taula_midi.tics_flanc[pos_midi - NUM_PRIMER_MIDI];
            now = rt_get_time();
            rt_task_make_periodic(&tasca_music[0], now, tics_flanc_nota*2);
            rt_task_make_periodic(&tasca_music[1], now+tics_flanc_nota, tics_flanc_nota*2);
         }
         else {
            rt_task_suspend(&tasca_music[0]);
            rt_task_suspend(&tasca_music[1]); 
         }
      } 
      else {
         now = rt_get_time();
         if (pos_midi != 0) {
            if (speaker_tancat) {
               speaker_tancat = 0;
               obrir_speaker();
            }
            tics_flanc_nota = taula_midi.tics_flanc[pos_midi - NUM_PRIMER_MIDI];
            rt_task_make_periodic(&tasca_music_speaker, now, INFINIT);
         }
         else {
            rt_task_suspend(&tasca_music_speaker);
            tancar_speaker();
            speaker_tancat = 1;
         }
      }
      
      pos_lectura_buffer++;
      now = rt_get_time();
      rt_task_make_periodic(&tasca_director, now+tics_duracio, INFINIT);
      
      rt_task_wait();
   }
   
}

/* -------------------------------------------------------------------------------- */
/*  manegador_notes: manegador de la RT-FIFO destinada a rebre un buffer de notes.  */
/*                   Utilitzant la tecnica del doble buffer, es guardara el buffer  */
/*                   rebut en el contrari al que actualment s'esta emprant.         */
/* -------------------------------------------------------------------------------- */

int manegador_notes(unsigned int fifo)
{
   int valor;
   int seguent_buffer;
   RTIME now;

   if (buffer_actual == 0) {
      seguent_buffer = 1;
   }
   else {
      seguent_buffer = 0;
   }
   valor = rtf_get(fifo, &buffers[seguent_buffer], sizeof(struct struct_missatge));
   if ((valor == sizeof(struct struct_missatge)) && (freq_midi_inicialitzat == 1)){
      if (inici_melodia == 1) {
         buffer_actual = 0;
         inici_melodia = 0;
         now = rt_get_time();
         obrir_speaker();
         rt_task_make_periodic(&tasca_director, now , INFINIT);
      }
      return 0;
   }
   else {
      return -EINVAL;
   } 
}

/* -------------------------------------------------------------------------- */
/*  manegador_frequencies_midi: manegador de la RT-FIFO encarregada de rebre  */
/*                              inicialment les frequencies i els tics dels   */
/*                              midis entre 50 i 100.                         */
/* -------------------------------------------------------------------------- */

int manegador_frequencies_midi(unsigned int fifo)
{
   int valor;

   valor = rtf_get(fifo, &taula_midi, sizeof(taula_midi));
   if (valor == sizeof(struct taula_musical)) {
      freq_midi_inicialitzat = 1;
      rtf_put(3,"MIDIS_OK\0",9);       // Confirmacio de correcta recepcio a RT-Linux
      return 0;
   }
   else {
      return -EINVAL;
   }
}

/* ----------------------------------------------------------- */
/*  init_module: funcio de inicialitzacio del modul rt_linux.  */
/* ----------------------------------------------------------- */

int init_module(void)
{
   inicialitzar_dades();
   rtf_create(1, TAM_TAULA_MIDI * 10);  // RT-FIFO entrada: frequencies midi inicials
   rtf_create(2, TAM_BUFFER * 10);      // RT-FIFO entrada: notes de la melodia
   rtf_create(3, 100);                  // RT-FIFO sortida: Notificacions a Linux
   rt_task_init(&tasca_music[0], music, 0, 3000, 4);
   rt_task_init(&tasca_music[1], music, 255, 3000, 3);
   rt_task_init(&tasca_music_speaker, music_speaker, 0, 3000, 4);
   rt_task_init(&tasca_director, director, 1, 3000, 5);
   rtf_create_handler(1, &manegador_frequencies_midi);
   rtf_create_handler(2, &manegador_notes);
   return 0;
}

/* -------------------------------------------------------------- */
/*  cleanup_module: funcio de inicialitzacio del modul rt_linux.  */
/* -------------------------------------------------------------- */

void cleanup_module(void)
{
   if (speaker) {
      rt_task_delete(&tasca_music_speaker);
   } 
   else {
      rt_task_delete(&tasca_music[0]);
      rt_task_delete(&tasca_music[1]);
   }
   rt_task_delete(&tasca_director);
   rtf_destroy(1);
   rtf_destroy(2);
   rtf_destroy(3);
}

